unit cliParameter;

{ The CLI parameter are:
      <input wav file> <options>

  output wav file name for input file name in.wav: in.lossy.wav

  options: -1 or -2 or -3        quality level (1 best, 3 worst, default: 2)
           -c <n>                forced blocksize (targeting specific codec behavior)
           -b <n>                forced output bitdepth (16 or 24)
           -o <output folder>    forced destination folder for the output file
           -flac                 optimized for use with FLAC
           -wv                   optimized for use with wavPack
           -tak                  optimized for use with TAK
}

interface

function getParms: boolean;           // interprets the cli parameter and stores them internally.
                                      // must precede any call to the other functions.
                                      // returns False in case there's something wrong with the options
function parmsError: string;          // returns an error text for the case that getParms returned False.
function wavIn: string;               // name of the input wav file
function wavOut: string;              // name of the output wav file based on the input file name.
                                      // with -d option used wavout is the complete path using the destination folder
function quality: integer;            // quality level. default: 2
function forcedBlocksize: integer;    // forced blocksize. 0 has the meaning 'no forced blocksize given'
function forcedOutBitDepth: integer;  // forced output bitdepth. 0 has the meaning 'no forced output bitdepth given'
function outFolder: string;           // forced folder for the output file. '' has the meaning 'no forced output folder given'
function outCodec: string;            // 'flac', 'wv', or 'tak' if the output file is to be optimized for one of these codecs.
                                      // '' has the meaning 'no codec given'.

implementation
uses
  SysUtils,
  wavIO;

var
    parmError, waveIn, folder, codec: string;
    qualityLevel, forceBlocksize, forceOutBitDepth: integer;
    NextParamStrParamNr: integer; // internal param counter for nextParamStr

procedure initNextParamStr;
begin
 NextParamStrParamNr := 0;
end;

function nextParamStr(var parm: string): boolean;
begin
 if NextParamStrParamNr = paramCount then
    begin
    nextParamStr := False;
    exit;
    end;
 NextParamStrParamNr := NextParamStrParamNr + 1;
 parm := ParamStr(NextParamStrParamNr);
 nextParamStr := True;
end;

function getParms: boolean;
var s: string;
    valError: integer;
begin
 getParms := False;
 waveIn := '';
 initNextParamStr;
 if not nextParamStr(s) then
    begin
    parmError := 'No input file given.';
    exit;
    end;
 if s[1] = '-' then
    begin
    parmError := 'No input file given.';
    exit;
    end;
 waveIn := s;
 qualityLevel := 0;
 forceBlocksize := 0;
 forceOutBitDepth := 0;
 folder := '';
 codec := '';
 while nextParamStr(s) do
    begin
    if s = '-1' then
       begin
       if qualityLevel <> 0 then
          begin
          parmError := 'Multiple quality level given.';
          exit;
          end;
       qualityLevel := 1;
       continue;
       end;
    if s = '-2'then
       begin
       if qualityLevel <> 0 then
          begin
          parmError := 'Multiple quality level given.';
          exit;
          end;
       qualityLevel := 2;
       continue;
       end;
    if s = '-3' then
       begin
       if qualityLevel <> 0 then
          begin
          parmError := 'Multiple quality level given.';
          exit;
          end;
       qualityLevel := 3;
       continue;
       end;
    if s = '-flac' then
       begin
       if codec <> '' then
          begin
          parmError := 'Multiple codecs given for which the output is to be optimized.';
          exit;
          end;
       codec := 'flac';
       continue;
       end;
    if s = '-wv' then
       begin
       if codec <> '' then
          begin
          parmError := 'Multiple codecs given for which the output is to be optimized.';
          exit;
          end;
       codec := 'wv';
       continue;
       end;
    if s = '-tak' then
       begin
       if codec <> '' then
          begin
          parmError := 'Multiple codecs given for which the output is to be optimized.';
          exit;
          end;
       codec := 'tak';
       continue;
       end;
    if s = '-b' then
       begin
       if forceOutBitDepth <> 0 then
          begin
          parmError := 'Multiple forced output bitdepth given.';
          exit;
          end;
       if not nextParamStr(s) then
          begin
          parmError := 'No forced output bitdepth given after -b.';
          exit;
          end;
       Val(s, forceOutBitDepth, valError);
       if valError <> 0 then
          begin
          parmError := 'Forced output bitdepth given is not an integer.';
          exit;
          end;
       if (forceOutBitDepth <> 16) and (forceOutBitDepth <> 24) then
          begin
          parmError := 'Only 16 or 24 allowed as forced output bitdepth.';
          exit;
          end;
       continue;
       end;
    if s = '-c' then
       begin
       if forceBlocksize <> 0 then
          begin
          parmError := 'Multiple forced blocksize given.';
          exit;
          end;
       if not nextParamStr(s) then
          begin
          parmError := 'No forced blocksize given after -c.';
          exit;
          end;
       Val(s, forceBlocksize, valError);
       if valError <> 0 then
          begin
          parmError := 'Forced blocksize given is not an integer.';
          exit;
          end;
       if forceBlocksize > maxblocksize then
          begin
          parmError := 'Max. forced blocksize allowed: ' + intToStr(maxblocksize);
          exit;
          end;
       continue;
       end;
    if s = '-o' then
       begin
       if folder <> '' then
          begin
          parmError := 'Multiple destination folder given.';
          exit;
          end;
       if not nextParamStr(s) then
          begin
          parmError := 'No destination folder given after -d.';
          exit;
          end;
       folder := s;
       if copy(folder, length(folder), 1) <> '\' then
          folder := folder + '\';
       continue;
       end;
    parmError := 'Incorrect option: ' + s;
    exit;
    end;
 if qualityLevel = 0 then
    qualityLevel := 2;
 getParms := True;
end;

function parmsError: string;
var newline: string;
begin
 newline := chr(13) + chr(10);
 parmsError := parmError + newline
             + newline
             + 'lossyWav usage: <input wav file> <options>' + newline
             + newline
             + 'options: -1 or -2 or -3        quality level (1 best, 3 worst, default: 2)' + newline
             + '         -c <n>                forced blocksize (targeting specific codec behavior)' + newline
             + '         -b <n>                forced output bitdepth (16 or 24)' + newline
             + '         -o <output folder>    forced destination folder for the output file' + newline
             + '         -flac                 optimized for use with FLAC' + newline
             + '         -wv                   optimized for use with wavPack' + newline
             + '         -tak                  optimized for use with TAK';
end;

function wavIn: string;
begin
 wavIn := waveIn;
end;

function wavOut: string;
begin
 if folder = '' then
    wavOut := copy(waveIn, 1, length(waveIn) - 3) + 'lossy.' + copy(waveIn, length(waveIn) - 2, 3)
 else
    wavOut := folder + copy(waveIn, 1, length(waveIn) - 3) + 'lossy.' + copy(waveIn, length(waveIn) - 2, 3)
end;

function quality: integer;
begin
 quality := qualityLevel;
end;

function forcedBlocksize: integer;
begin
 forcedBlocksize := forceBlocksize;
end;

function forcedOutBitDepth: integer;
begin
 forcedOutBitDepth := forceOutBitDepth;
end;

function outFolder: string;
begin
 outFolder := folder;
end;

function outCodec: string;
begin
 outCodec := codec;
end;

end.
