unit wavIO;

{ wavIO specializes on the situation where an input wav File is transformed into an output wav file
  with an identical structure but potentially different sample values.
  Reading and writing of sample values is done on a logical block basis (blocksize is configurable and as used in our context is roughly 1000 samples).
  Sample values are always 32 bit integers. With 16 (24) bit files the 16 (24) bit make up for the 16 (24) MSBs amd the remaining bit are set to 0.
  Reading and writing is buffered internally - this is just an internal detail not to be confused with the
  provided reading and writing functions which work on a logical sample block basis.
}

interface

const maxchannels = 8;
      maxblocksize = 32768;

type PLongInt = ^LongInt;
     allChannelsSampleBlock = array[0..maxchannels-1, 0..maxblocksize-1] of LongInt;


var
  infile : file;

function openWavIO(inWav: string; outWav: string; blocksize: LongInt; overlapsize: LongInt): boolean;
                                                            // False if operation not succesful.
                                                            // Reads the header chunks of inWav and memorizes the
                                                            // information from the fmt chunk.
                                                            // Creates outWav and fills its header chunks with an
                                                            // identical copy of that of inWav.
                                                            // Memorizes blocksize (nr of samples to read/write with subsequent calls to
                                                            // readNextSampleBlock resp. writeNextSampleBlock).
                                                            // Memorizes overlapsize (nr of last samples from previous readNextSampleBlock call to
                                                            // be put ahead of new samples fetched by readNextSampleBlock).
                                                            // overlapsize must be smaller than blocksize.
function channelCount: integer;                             // number of channels
function sampleRate: LongInt;                               // sample rate in Hz
function bitsPerSample: integer;                            // bits per sample
function readNextSampleBlock(var sampleBlock: allChannelsSampleBlock; var readsamplesCount: integer): boolean;
                                                            // False if not possible (usually because all the samples have been read).
                                                            // sampleBlock: next block of samples from input wav file for all the channels.
                                                            //              first sample is in element with index 0.
                                                            //              In case of overlapsize > 0 when using openWavIO the first overlapsize
                                                            //              samples are re-read according to previous readNextSampleBlock and
                                                            //              then followed by the readsamplesCount number of samples from the
                                                            //              input wav file.
                                                            // readsamplesCount: number of samples read (in each of the channels).
                                                            // In case of overlapsize > 0 when using openWavIO readsamplesCount does not count
                                                            // the overlap samples. readsamplesCount counts the new wave samples.
function writeNextSampleBlock(var sampleBlock: allChannelsSampleBlock; samplesCountToBeWritten: integer): boolean;
                                                            // False if not possible.
                                                            // Writes sampleBlock as the next sample block of all the channels to the output file.
                                                            // The number of samples to be written is given by samplesCountToBeWritten.
                                                            // The overlap area in sampleBlock is not written:
                                                            // First sample written is sampleBlock element with index overlapsize
                                                            // - overlapsize from openWavIO call.
                                                            // Last sample written is in element with index samplesCountToBeWritten - 1.
                                                            // samplesCountToBeWritten should be the readsamplesCount value from the
                                                            // corresponding readNextSampleBlock function call.
function closeWavIO: boolean;                               // False if not possible.
                                                            // Finalizes the wavIO operations.



implementation
uses
  SysUtils;

type PByte = ^Byte;
     PWord = ^Word;
     PSmallInt = ^SmallInt;

var outfile : File;
    filesOpened: boolean;
    inBuf, outBuf: PByte;
    overlapSampleSize, overlapBufSize: LongInt;
    bufsize: LongInt; // number of byte that make up for inbuf and outbuf
    nrOfByteInOneBlockInBuf: LongInt;  // number of byte that make up a complete block in inBuf and outBuf.
    bufferAllocated: boolean;
    nrOfBlocksInBuf: integer;  // inBuf size as a factor of blocksize
    nrOfFullBlocksReadIntoInBuf: integer;  // with the last BlockRead: how many full blocks managed it into inBuf ?
    nrOfByteOfTheLastIncompleteBlock: LongInt; // when EOF was encountered with the last BlockRead: the number of byte that didn't make it
                                               // into the last full block
    nrOfBlockInBufNotYetFetched: integer; // the next number of the block in inBuf that was not fetched yet by readNextSampleBlock.
                                          // 0: no data in inBuf;
    nrOfByteFilledInOutBuf: LongInt;      // how many byte have managed it into outBuf that are not written to the outputfile yet?
    chCount: integer;
    samplingrate: LongInt;
    bitdepth : integer;

function openWavIO(inWav: string; outWav: string; blocksize: LongInt; overlapsize: LongInt): boolean; // only standard chunks assumed so far
var
 buf : array[1..44] of char;
 pbuf, pfmt, p: PChar;
 readbyteCount : LongInt;
 begin
 openWavIO := False;
 filesOpened := False;
 bufferAllocated := false;
 {I-}
 AssignFile(infile, inWav);
 if IOResult <> 0 then
    exit;
 Reset(infile, 1);
 if IOResult <> 0 then
    exit;
 AssignFile(outfile, outwav);
 if IOResult <> 0 then
    exit;
 Rewrite(outfile, 1);
 if IOResult <> 0 then
    exit;
 filesOpened := True;
 BlockRead(infile, buf, 44, readbyteCount);
 if (IOResult <> 0) or (readbyteCount <> 44) then
    exit;
 BlockWrite(outfile, buf, 44);
 if IOResult <> 0 then
    exit;
 {I+}
 chCount := 0;
 samplingrate := 0;
 bitdepth := 0;
 if copy(buf, 1, 4) <> 'RIFF' then
    begin
    writeln('No RIFF file.');
    closeWavIO;
    exit;
    end;
 if copy(buf, 9, 4) <> 'WAVE' then
    begin
    writeln('No WAVE file.');
    closeWavIO;
    exit;
    end;
 if copy(buf, 13, 4) <> 'fmt ' then
    begin
    writeln('No instant fmt chunk in wav header.');
    closeWavIO;
    exit;
    end;
 if copy(buf, 37, 4) <> 'data' then
    begin
    writeln('data chunk not immediately following fmt chunk in wav header.');
    closeWavIO;
    exit;
    end;
 pbuf := @buf[1];
 pfmt := pbuf;
 Inc(pfmt, 12);   // pfmt points to fmt chunk
 p := pfmt;
 Inc(p, 8);       // points to data format in fmt chunk. content should be 1 (PCM format)
 if PWord(p)^ <> 1 then
    begin
    writeln('data chunk not immediately following fmt chunk in wav header.');
    closeWavIO;
    exit;
    end;
 p := pfmt;
 Inc(p, 10);      // points to number of channels in fmt chunk. content should be 1 (PCM format)
 chCount := PWord(p)^;
 p := pfmt;
 Inc(p, 12);      // points to sample rate in fmt chunk. content should be 1 (PCM format)
 samplingrate := PLongInt(p)^;
 p := pfmt;
 Inc(p, 22);      // points to bits per sample in fmt chunk. content should be 1 (PCM format)
 bitdepth := PWord(p)^;
 if (bitdepth <> 16) and (bitdepth <> 24) then
    begin
    writeln('Input file has a bitrate of ' + IntToStr(bitdepth) + ' bit per sample.');
    writeln('This is not supported.');
    closeWavIO;
    exit;
    end;
 if blocksize < 20000 then
    nrOfBlocksInBuf := 8
 else
    if blocksize < 40000 then
       nrOfBlocksInBuf := 4
    else
       nrOfBlocksInBuf := 1;
 nrOfByteInOneBlockInBuf := (bitdepth DIV 8) * chCount * blocksize;
 bufsize := nrOfByteInOneBlockInBuf * nrOfBlocksInBuf;
 GetMem(inBuf, bufsize);
 GetMem(outBuf, bufsize);
 overlapSampleSize := overlapsize;
 overlapBufSize := (bitdepth DIV 8) * chCount * overlapsize;
 bufferAllocated := True;
 {I-}
 BlockRead(infile, inbuf^, bufsize, readbyteCount);  // pre-read into inbuf
 if IOResult <> 0 then
    exit;
 {I+}
 if readbyteCount = 0 then
    exit; // no data
 nrOfFullBlocksReadIntoInBuf := readbyteCount DIV nrOfByteInOneBlockInBuf;
 nrOfByteOfTheLastIncompleteBlock := readbyteCount - nrOfFullBlocksReadIntoInBuf * nrOfByteInOneBlockInBuf;
 nrOfBlockInBufNotYetFetched := 1;
 nrOfByteFilledInOutBuf := 0;
 openWavIO := True;
 end;

function channelCount: integer;
 begin
   channelCount := chCount;
 end;

function sampleRate: LongInt;
 begin
   sampleRate := samplingrate;
 end;

function bitsPerSample: integer;
 begin
   bitsPerSample := bitdepth;
 end;

procedure transfer(var sampleBlock: allChannelsSampleBlock; sampleBlockTransferStart: LongInt; pStartB: PByte; pEndB: PByte);
 var
 pInt16, pEndInt16: PSmallInt;
 pB: PByte;
 loByte: Byte;
 i, iChannel: integer;
 begin
 i := sampleBlockTransferStart;
 if bitdepth = 16 then
    begin
    pInt16 := PSmallInt(pStartB);
    pEndInt16 := PSmallInt(pEndB);
    while pInt16 <> pEndInt16 do
       begin
       for iChannel := 0 to chCount - 1 do
           begin
           sampleBlock[iChannel, i]:= 65536 * LongInt(pInt16^);
           Inc(pInt16);
           end;
       i := i + 1;
       end;
    end
 else  // 24 bit
    begin
    pB := pStartB;
    while pB <> pEndB do
       begin
       for iChannel := 0 to chCount - 1 do
           begin
           loByte := pB^;
           Inc(pB);
           pInt16 := PSmallInt(pB);
           Inc(pB, 2);
           sampleBlock[iChannel, i] := 65536 * pInt16^ + 256 * loByte;
           end;
       i := i + 1;
       end;
    end;
 end;

function readNextSampleBlock(var sampleBlock: allChannelsSampleBlock; var readsamplesCount: integer): boolean;
 var
 readbyteCount, sampleBlockStart: LongInt;
 pB, pStartB: PByte;
 lastNrOfBlockInBufNotYetFetched: integer;
 begin
   readNextSampleBlock := False;
   sampleBlockStart := overlapSampleSize;
   if nrOfBlockInBufNotYetFetched = 0 then // samples to be fetched aren't in inbuf
      begin
      if nrOfFullBlocksReadIntoInBuf < nrOfBlocksInBuf then // EOF encountered with last call to readNextSampleBlock
         exit; // no more data
      sampleBlockStart := overlapSampleSize;
      if overlapSampleSize > 0 then
         begin
         pB := inBuf;
         Inc(pB, nrOfFullBlocksReadIntoInBuf * nrOfByteInOneBlockInBuf);
         pStartB := pB;
         Dec(pStartB, overlapBufSize);
         transfer(sampleBlock, 0, pStartB, pB);
         end;
      {I-}
      BlockRead(infile, inbuf^, bufsize, readbyteCount);
      if IOResult <> 0 then
         exit;
      {I+}
      if readbyteCount = 0 then
         exit; // no more data
      nrOfFullBlocksReadIntoInBuf := readbyteCount DIV nrOfByteInOneBlockInBuf;
      nrOfByteOfTheLastIncompleteBlock := readbyteCount - nrOfFullBlocksReadIntoInBuf * nrOfByteInOneBlockInBuf;
      nrOfBlockInBufNotYetFetched := 1;
      end;
   pB := inBuf;
   Inc(pB, (nrOfBlockInBufNotYetFetched - 1) * nrOfByteInOneBlockInBuf);
   pStartB := pB; // points to first sample in Block
   lastNrOfBlockInBufNotYetFetched := nrOfBlockInBufNotYetFetched;
   if nrOfBlockInBufNotYetFetched <= nrOfFullBlocksReadIntoInBuf then
      begin
      Inc(pB, nrOfByteInOneBlockInBuf);  // pB points to the spot after last sample in Block
      readsamplesCount := nrOfByteInOneBlockInBuf DIV ((bitdepth DIV 8) * chCount);
      nrOfBlockInBufNotYetFetched := nrOfBlockInBufNotYetFetched + 1;
      if nrOfBlockInBufNotYetFetched > nrOfBlocksInBuf then
         nrOfBlockInBufNotYetFetched := 0;
      end
   else
      begin
      if nrOfByteOfTheLastIncompleteBlock = 0 then
         exit; // no more data
      Inc(pB, nrOfByteOfTheLastIncompleteBlock);  // pB points to the spot after last sample in Block
      readsamplesCount := nrOfByteOfTheLastIncompleteBlock DIV ((bitdepth DIV 8) * chCount);
      nrOfBlockInBufNotYetFetched := 0; // force a try for a new Blockread with next call which will lead to a return False
      end;
   if lastNrOfBlockInBufNotYetFetched > 1 then
      begin
      if overlapSampleSize > 0 then
         Dec(pStartB, overlapBufSize);
      sampleBlockStart := 0;
      end;
   transfer(sampleBlock, sampleBlockStart, pStartB, pB);
   readNextSampleBlock := True;
 end;

procedure Int24ToInt16andLoByte(int24: LongInt; var hiInt16: SmallInt; var loByte: Byte);
// done in such a way that: 24bit = 256 * hiInt + loByte such that loByte >= 0;
var
 loByteTest: SmallInt;
begin
 hiInt16 := int24 DIV 256;
 loByteTest := int24 - 256 * hiInt16;
 if loByteTest >= 0 then
    loByte := Byte(loByteTest)
 else
    begin
    hiInt16 := hiInt16 - 1;
    loByte := Byte(loByteTest + 256);
    end;
end;

function writeNextSampleBlock(var sampleBlock: allChannelsSampleBlock; samplesCountToBeWritten: integer): boolean;
var
 nrOfByteForOutBuf, sample: LongInt;
 pB: PByte;
 pInt16: PSmallInt;
 i, iChannel: integer;
 sampleLoByte: Byte;
 sampleHiInt: SmallInt;
 begin
   writeNextSampleBlock := False;
   nrOfByteForOutBuf := (bitdepth DIV 8) * chCount * samplesCountToBeWritten;
   if nrOfByteFilledInOutBuf + nrOfByteForOutBuf > bufsize then // not sufficient space in outBut: flush outBuf
      begin
      {I-}
      BlockWrite(outfile, outBuf^, nrOfByteFilledInOutBuf);
      if IOResult <> 0 then
         exit;
      {I+}
      nrOfByteFilledInOutBuf := 0;
      end;
   pB := outBuf;
   Inc(pB, nrOfByteFilledInOutBuf); // points to first free spot in outBuf
   if bitdepth = 16 then
      begin
      pInt16 := PSmallInt(pB);
      for i:= overlapSampleSize to overlapSampleSize + samplesCountToBeWritten - 1 do
          for iChannel := 0 to chCount - 1 do
              begin
              pInt16^ := sampleBlock[iChannel, i] DIV 65536;
              Inc(pInt16);
              end
      end
   else // 24 bit
      for i:= overlapSampleSize to overlapSampleSize + samplesCountToBeWritten - 1 do
          for iChannel := 0 to chCount - 1 do
              begin
              sample := sampleBlock[iChannel, i] DIV 256; // 24 bit value
              Int24ToInt16andLoByte(sample, sampleHiInt, sampleLoByte);
              pB^ := sampleLoByte;
              Inc(pB);
              pInt16 := PSmallInt(pB);
              pInt16^ := sampleHiInt;
              Inc(pB, 2);
              end;
   nrOfByteFilledInOutBuf := nrOfByteFilledInOutBuf + nrOfByteForOutBuf;
   writeNextSampleBlock := True;
 end;

function closeWavIO: boolean;
 begin
   closeWavIO := False;
   if filesOpened then
      begin
      if nrOfByteFilledInOutBuf > 0 then
         begin
         {I-}
         BlockWrite(outfile, outBuf^, nrOfByteFilledInOutBuf);
         if IOResult <> 0 then
            exit;
         {I+}
         end;
      {I-}
      closeFile(infile);
      if IOResult <> 0 then
         exit;
      CloseFile(outfile);
      if IOResult <> 0 then
         exit;
      {I+}
      filesOpened := False;
   end;
   if bufferAllocated then
      begin
      ReAllocMem(inBuf, 0);
      ReAllocMem(outBuf, 0);
      overlapSampleSize := 0;
      overlapBufSize := 0;
      bufferAllocated := False;
      end;
   closeWavIO := True;
 end;

end.
